#include <boost/concept_check.hpp>
#include <opencv/cv.h>

enum MATCHING_METHOD
{
    SED=1,NCC=2,Census=3
};

class FlowField
{
public:
    enum consitencyCheck
    {
        //Allows to configure outlier filtering
        CONSI_NONE = 0, CONSI_STD = 0b111,  CONSI_2X  = 0b011,  CONSI_1X  = 0b001, CONSI_REGION = 0b100
    };
    cv::Mat1f _filterWeights;

protected:

    cv::Mat2f _fields[4];
    int _r, _r2,_scales,_n;
    cv::Mat _from, _to;
    cv::Mat _fromFeature, _toFeature;
    int _method;

    static cv::Mat2f propagateScale ( int fscale, int psize, cv::Mat img1, cv::Mat img2, int method = Census, cv::Mat2f inflow=cv::Mat2f(), cv::Mat img1feature=cv::Mat(), cv::Mat img2feature = cv::Mat() );
    void forwardBackwardInconsisitencies ( cv::Mat2f& f1, cv::Mat2f f2, float threshold, cv::Mat1f weight );
    void regionFilter ( cv::Mat2f flow, cv::Mat2f flowOrg, float minRegionSize = 50.f,   float maxdif = 3.f ); // maxdif = 0 deactivates check;
    std::vector< float > sparseMatchesList ( cv::Mat2f flow,  cv::Mat1f weights,  int localDensity = 4 );

public:

    /** Allows to generate a raw (unfiltered) Flow Field without instancing the class: */
    static cv::Mat2f genFlowField ( cv::Mat from , cv::Mat to,  int r, int n, int method = Census, cv::Mat fromFeature = cv::Mat(), cv::Mat toFeature = cv::Mat() );

    /** Constructor. Input images must be provided in the "from" and "to" parameters. If feature based images are used (e.g. SIFTFlow ) they must be  provided in the parameters "fromFeature" and "toFeatue". */
    FlowField ( cv::Mat from, cv::Mat to, int r, int r2, int method = Census , cv::Mat fromFeature = cv::Mat(), cv::Mat toFeature  = cv::Mat(), int hierarchies = 3 );

    /** Generates and returns outlier filtered Flow Field. Also generates backward flows if not existing.  filterThreshold is "epsilon" in paper,  minRegionSize  is "s" in paper */
    cv::Mat2f  genFiltered ( float filterThreshold=5.f, float minRegionSize = 50.f,  int  cc = CONSI_STD );

    /** Generates and returns matches for EpicFlow (see Section 3.5 in paper). Every match consist of 4 consecutive floats (x1,y1,x2,y2) .
        You have to call genFiltered() before calling this function! localDensity is "e" in paper */
    std::vector< float >  getEpicMatches ( int localDensity = 4 );

    /** Returns internally saved flow fields. 0 = outlier filtered flow field. 1 = forward flow (unfiltered flow field), 2 = backward flow with patch radius r, 3 = backward flow with patch radius r_2
        The parameter 1 is always valid. 0,2,3 are valid after calling ff.genFiltered() at least once. */
    inline cv::Mat2f & getFlowField ( int which = 1 )
    {
        return _fields[which];
    }
};
