#include "opencv2/core/core.hpp"
#include "opencv2/highgui/highgui.hpp"

#include <fstream>
#define _cvMatf(y) cv::Mat_ < cv::Vec< float, y> >

#ifdef ENABLE_SIFTFLOW

#include "opencv2/features2d/features2d.hpp"
#include <opencv2/xfeatures2d/nonfree.hpp>

_cvMatf ( 128 )  SIFT_ ( cv::Mat1b bw, float desSize = 0.5f )
{
    _cvMatf ( 128 ) res ( bw.rows,bw.cols );
    cv::Mat1f descritors;
    cv::Ptr< cv::DescriptorExtractor>  de = cv::xfeatures2d::SIFT::create();//new cv::SiftDescriptorExtractor();
    std::vector< cv::KeyPoint > kpv;

    for ( int i=0; i<bw.rows; i++ ) for ( int j=0; j<bw.cols; j++ )
            kpv.push_back ( cv::KeyPoint ( j,i,desSize,0 ) );

    cv::Mat1f descriptors ( bw.cols*bw.rows, 128 );
    de->compute ( bw,kpv,descriptors );

    for ( int i=0; i<bw.rows; i++ ) for ( int j=0; j<bw.cols; j++ )
        {
            int ix = i* ( bw.cols );
            for ( int k =0; k<descriptors.cols; k++ ) res ( i,j ) [k] = descriptors ( ix+j,k );
        }
    return res;
}

template <int components, int inputChannels>
void PCA ( _cvMatf ( inputChannels ) in,_cvMatf ( inputChannels ) in2, _cvMatf ( components ) & out,_cvMatf ( components ) & out2, int ps = 0 )
{
    _cvMatf ( inputChannels ) inb;
    _cvMatf ( inputChannels ) in2b;

    if ( ps )
    {
        cv::copyMakeBorder ( in,inb,ps,ps,ps,ps,cv::BORDER_REPLICATE );
        cv::copyMakeBorder ( in2,in2b,ps,ps,ps,ps,cv::BORDER_REPLICATE );
    }
    else
    {
        inb = in;
        in2b = in2;
    }
    int psd = ps*2+1;
    assert ( psd*psd*inputChannels >= components );
    int gsize =  in.cols  *in.rows;

    out = _cvMatf ( components ) ( in2.rows,in2.cols );
    out2 = _cvMatf ( components ) ( in2.rows,in2.cols );
    cv::Mat1f feat ( gsize*2, psd*psd* in.channels() );
    cv::Mat1f res;


    for ( int i=0; i<in.rows; i++ ) for ( int j=0; j<in.cols; j++ )
        {
            int ix = i*in.cols;
            for ( int k=0; k<psd; k++ ) for ( int l=0; l<psd; l++ )
                {
                    int klx = ( k*psd+l ) *in.channels();
                    cv::Vec< float, inputChannels>   & t =  inb ( i+k,j+l );
                    for ( int m =0; m<in.channels(); m++ )
                        feat ( ix+j, klx+m ) = t[m];
                }
        }
    for ( int i=0; i<in.rows; i++ ) for ( int j=0; j<in.cols; j++ )
        {
            int ix = i*in.cols + gsize;

            for ( int k=0; k<psd; k++ ) for ( int l=0; l<psd; l++ )
                {
                    int klx = ( k*psd+l ) *in.channels();
                    cv::Vec< float, inputChannels>     & t =  in2b ( i+k,j+l );
                    for ( int m =0; m<in.channels(); m++ )
                        feat ( ix+j, klx+m ) = t[m];
                }
        }
    cv::PCA pca ( feat,cv::noArray(),CV_PCA_DATA_AS_ROW,components );
    pca.project ( feat,res );

    assert ( res.cols == components );
    for ( int i=0; i<in.rows; i++ ) for ( int j=0; j<in.cols; j++ )
        {
            int ix = i* ( in.cols );
            for ( int k =0; k<components; k++ ) out ( i,j ) [k] = res ( ix+j,k );
            int ix2 = i* ( in.cols ) + gsize;
            for ( int k =0; k<components; k++ ) out2 ( i,j ) [k] = res ( ix2+j,k );
        }
}
#endif

void  saveFlowField ( const cv::Mat2f & flow, std::string filename )
{
    std::ofstream file;
    file.open ( filename,  std::ios::out |  std::ios::binary );
    assert ( file.is_open() );

    float x=202021.25;
    int w = flow.cols;
    int h = flow.rows;

    file.write ( ( char* ) &x,sizeof ( float ) );
    file.write ( ( char* ) &w,sizeof ( int ) );
    file.write ( ( char* ) &h,sizeof ( int ) );

    for ( int i =0; i<h; i++ ) for ( int j =0; j<w; j++ )
        {
            file.write ( ( char* ) &flow ( i,j ) [0],sizeof ( float ) );
            file.write ( ( char* ) &flow ( i,j ) [1],sizeof ( float ) );
        }
}


void saveMatches ( std::vector<float> matches, std::string filename )
{
    std::ofstream file ( filename );
    assert ( file.is_open() );
    assert ( matches.size() %4==0 );

    for ( int i =0; i<matches.size(); i++ )
    {
        file << matches[i] << " ";
        if ( i%4==3 ) file << "\n";
    }
    file.close();
}
